#include "libsangoma_freetdm.h"

/*==========================================
 * Function Prototypes
 *=========================================*/
static int sangoma_freetdm_gsm_create_span(sangoma_port_obj_t port, ftdm_span_t * span);
static int sangoma_freetdm_gsm_config(sangoma_port_obj_t port, ftdm_span_t * span, span_config_t * span_conf);

static int sangoma_freetdm_tapping_create_pri_span(sangoma_port_obj_t port, ftdm_span_t * span);
static int sangoma_freetdm_tapping_create_bri_span(sangoma_port_obj_t port, ftdm_span_t * span);

static int sangoma_freetdm_tapping_create_isup_sig_span(sangoma_port_obj_t port, ftdm_span_t * span);
static int sangoma_freetdm_tapping_create_isup_pure_voice_span(sangoma_port_obj_t port, ftdm_span_t * span);

static int sangoma_freetdm_tapping_config(sangoma_port_obj_t port, ftdm_span_t * span, span_config_t * span_conf);
static int sangoma_freetdm_isup_tapping_config(sangoma_port_obj_t port, ftdm_span_t * span, span_config_t * span_conf);
static int sangoma_freetdm_analog_create_span(sangoma_port_obj_t port, ftdm_span_t * span);
static int sangoma_freetdm_analog_config(sangoma_port_obj_t port, ftdm_span_t * span, span_config_t * span_conf);

static int sangoma_freetdm_mtp2_api_create_span(sangoma_port_obj_t port, ftdm_span_t * span);
static int sangoma_freetdm_mtp2_api_config(sangoma_port_obj_t port, ftdm_span_t * span, span_config_t * span_conf);

/*==========================================
 * Globals
 *=========================================*/

/* Global user supplied information */
static sangoma_freetdm_init_t g_ftdm_init;

int sangoma_freetdm_init(sangoma_freetdm_init_t * ftdm_init)
{

	if (!ftdm_init) {
		fprintf(stderr, "%s: Error invalid freetdm initialization struct\n", __FUNCTION__);
		return -1;
	}

	if (!ftdm_init->event_msg_queue_sz) {
		ftdm_init->event_msg_queue_sz = MAX_QUEUE_SIZE;
	}

	/* SET Logger Details
	 * if logger is NULL then set libsangoma freetdm default logger
	 * else set user application logger as requested */
	if (ftdm_init->logger == NULL ) {
		ftdm_global_set_logger(sangoma_freetdm_logger);
	} else {
		ftdm_global_set_logger(ftdm_init->logger);
	}

	ftdm_global_set_crash_policy(ftdm_init->crash_policy_level);

	/* Freetdm global initialization. 
	 * Start using ftdm_log ftdm initialization 
	 */
	if (ftdm_global_init() != FTDM_SUCCESS) {
		fprintf(stderr, "Error loading FreeTDM\n");
		return -1;
	}

	/* create global queue */
	if (ftdm_queue_create(&ftdm_init->event_msg_queue, ftdm_init->event_msg_queue_sz) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to create global event queue. Exit!\n");
		return -1;
	}

	memcpy(&g_ftdm_init, ftdm_init, sizeof(g_ftdm_init));

	return 0;
}

int sangoma_freetdm_destroy(void)
{
	ftdm_queue_destroy(&g_ftdm_init.event_msg_queue);
	return ftdm_global_destroy();
}

/* FreeTDM Event call back function
 * User is not allowed to execute FreeTDM API calls here
 * As it can lead to deadlock.
 * We copy the event msg and send to application event
 * handling thread.
 */
FIO_SIGNAL_CB_FUNCTION(on_freetdm_event_signal)
{
	ftdm_sigmsg_t *sig_ev;
	sig_ev = ftdm_malloc(sizeof(*sigmsg));
	memcpy(sig_ev, sigmsg, sizeof(*sigmsg));
	if (sigmsg->raw.data)
	{
		sig_ev->raw.data = ftdm_malloc(sigmsg->raw.len);
		memcpy(sig_ev->raw.data, sigmsg->raw.data, sigmsg->raw.len);
	}
	ftdm_queue_enqueue(g_ftdm_init.event_msg_queue, sig_ev);
	return FTDM_SUCCESS;
}

/* FreeTDM genertic create spane.
 * Public function used by application.
 * Detect confguratoin based on sangoma hw port object
 */
int sangoma_freetdm_create_span(sangoma_port_obj_t port, ftdm_span_t ** span, int operating_mode)
{
	/* Initializing local variables */
	uint32_t configured = 0;
	char ftdm_name[255];
	ftdm_span_t *ftdm_span = NULL;
	int port_id = sangoma_config_port_get_port_id(port);
	int card_type = sangoma_config_port_hwinfo_get_model_id(port);

	*span = NULL;

	sprintf(ftdm_name, "wp%d", port_id);

	/* Create and configure Span depedning upon the wanpipe configured */
	if (ftdm_span_create("wanpipe", ftdm_name, &ftdm_span) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "unable to create span %s of type GSM\n", ftdm_name);
		return -1;
	}

	ftdm_log(FTDM_LOG_DEBUG, "Successfully created span %s of type %s\n",
			 ftdm_name, sangoma_config_port_hwinfo_get_model_str(port));

	switch (card_type) {
	case AFT_ADPTR_W400:
		configured = sangoma_freetdm_gsm_create_span(port, ftdm_span);
		break;

		/* Adding cardtype supported for pri tapping solution */
	case A101_ADPTR_2TE1:
	case A104_ADPTR_4TE1:
	case A108_ADPTR_8TE1:
	case A116_ADPTR_16TE1:
		if (operating_mode == SNG_ISDN_PRI_TAPPING_MODE) {
			configured = sangoma_freetdm_tapping_create_pri_span(port, ftdm_span);
		} else if (operating_mode == SNG_SS7_ISUP_SIG_TAPPING_MODE) {
			configured = sangoma_freetdm_tapping_create_isup_sig_span(port, ftdm_span);
		} else if (operating_mode == SNG_SS7_ISUP_PURE_VOICE_TAPPING_MODE) {
			configured = sangoma_freetdm_tapping_create_isup_pure_voice_span(port, ftdm_span);
		} else if ((operating_mode == SNG_SS7_MTP2_API_MODE_LSL) || (operating_mode == SNG_SS7_MTP2_API_MODE_HSL)){
			configured = sangoma_freetdm_mtp2_api_create_span(port, ftdm_span);
		} else {
			ftdm_log(FTDM_LOG_ERROR, "Error: Unsupported Sangoma hw type %s\n",
					sangoma_config_port_hwinfo_get_model_str(port));
			return -1;
		}
		break;

		/* Adding cardtype supported for bri tapping solution */
	case AFT_ADPTR_ISDN:
		configured = sangoma_freetdm_tapping_create_bri_span(port, ftdm_span);
		break;

	case A200_ADPTR_ANALOG:
	case A400_ADPTR_ANALOG:
		configured = sangoma_freetdm_analog_create_span(port, ftdm_span);
		break;
	default:
		ftdm_log(FTDM_LOG_ERROR, "Error: Unsupported Sangoma hw type %s\n",
				 sangoma_config_port_hwinfo_get_model_str(port));
		return -1;
	}

	ftdm_log(FTDM_LOG_DEBUG, "Total configured channels %u\n", configured);
	if (!configured) {
		return -1;
	}

	*span = ftdm_span;

	sangoma_msleep(2000);

	return 0;
}

int sangoma_freetdm_span_config_and_start(sangoma_port_obj_t port, ftdm_span_t * ftdm_span, span_config_t * span_conf, int operating_mode)
{
	int card_type = sangoma_config_port_hwinfo_get_model_id(port);
	int err = -1;

	/* Filter out all other cards just look for W400 */
	switch (card_type) {
	case AFT_ADPTR_W400:
		err = sangoma_freetdm_gsm_config(port, ftdm_span, span_conf);
		if (err) {
			return -1;
		}
		ftdm_log(FTDM_LOG_DEBUG, "Configured GSM FreeTDM span %d\n", ftdm_span_get_id(ftdm_span));
		break;
	case A101_ADPTR_2TE1:
	case A104_ADPTR_4TE1:
	case A108_ADPTR_8TE1:
	case A116_ADPTR_16TE1:
	case AFT_ADPTR_ISDN:
		if ((operating_mode == SNG_ISDN_PRI_TAPPING_MODE) || (operating_mode == SNG_ISDN_BRI_TAPPING_MODE)) {
			err = sangoma_freetdm_tapping_config(port, ftdm_span, span_conf);
			ftdm_log(FTDM_LOG_DEBUG, "Configured ISDN Tapping FreeTDM span %d\n", ftdm_span_get_id(ftdm_span));
		} else if ((operating_mode == SNG_SS7_ISUP_SIG_TAPPING_MODE) && (card_type != AFT_ADPTR_ISDN)) {
			err = sangoma_freetdm_isup_tapping_config(port, ftdm_span, span_conf);
			ftdm_log(FTDM_LOG_DEBUG, "Configured ISUP Tapping FreeTDM span %d\n", ftdm_span_get_id(ftdm_span));
		} else if (((operating_mode == SNG_SS7_MTP2_API_MODE_LSL) || (operating_mode == SNG_SS7_MTP2_API_MODE_HSL)) &&
				(card_type != AFT_ADPTR_ISDN)) {
			err = sangoma_freetdm_mtp2_api_config(port, ftdm_span, span_conf);
			ftdm_log(FTDM_LOG_DEBUG, "Configured MTP2 API FreeTDM span %d\n", ftdm_span_get_id(ftdm_span));
		} else {
			ftdm_log(FTDM_LOG_ERROR, "Error: Unsupported Tapping mode %d by hwtype %s\n",
					operating_mode, sangoma_config_port_hwinfo_get_model_str(port));
			return -1;
		}
		if (err) {
			return -1;
		}
		break;

	case A200_ADPTR_ANALOG:
	case A400_ADPTR_ANALOG:
		err = sangoma_freetdm_analog_config(port, ftdm_span, span_conf);
		if (err) {
			return -1;
		}
		ftdm_log(FTDM_LOG_DEBUG, "Configured Analog FreeTDM span %d\n", ftdm_span_get_id(ftdm_span));
		break;
	default:
		ftdm_log(FTDM_LOG_ERROR, "Error: Unsupported Sangoma hw type %s\n",
				 sangoma_config_port_hwinfo_get_model_str(port));
		return -1;
	}

	/* Start the configured span */
	return ftdm_span_start(ftdm_span);
}

int sangoma_freetdm_send_sms(ftdm_span_t * span, char *called, char *text)
{
	char *rply;
	char cmd[1024];

	sprintf(cmd, "gsm sms %s %s %s", ftdm_span_get_name(span), called, text);

	ftdm_log(FTDM_LOG_DEBUG, "\nSMS cmd: %s\n", cmd);

	rply = ftdm_api_execute(cmd);

	if (rply) {
		ftdm_log(FTDM_LOG_DEBUG, "%s", rply);
		ftdm_free(rply);
	}

	return 0;
}
int sangoma_freetdm_send_at(ftdm_span_t * span, char *command)
{
	char *rply;
	char cmd[1024];

	sprintf(cmd, "gsm exec %s %s", ftdm_span_get_name(span), command);
	
	ftdm_log(FTDM_LOG_DEBUG, "\nExecuting GSM EXEC cmd: '%s' on span %s\n", command, ftdm_span_get_name(span));

	rply = ftdm_api_execute(cmd);
	if (rply) {
		ftdm_log(FTDM_LOG_DEBUG, "%s", rply);
		ftdm_free(rply);
	}

	return 0;
}

int sangoma_freetdm_call_req(ftdm_span_t * ftdm_span, char *called, char *calling)
{
	unsigned int span_id = 0;
	ftdm_caller_data_t call_data;
	ftdm_hunting_scheme_t hunting;

	/*
	   ftdm_assert_return(called, -1, "Invalid called number\n");
	   ftdm_assert_return(calling, -1, "Invalid calling number\n");
	 */

	memset(&call_data, 0, sizeof(call_data));

	span_id = ftdm_span_get_id(ftdm_span);

	/* set up hunting mode data */
	hunting.mode = FTDM_HUNT_CHAN;
	hunting.mode_data.chan.span_id = span_id;
	hunting.mode_data.chan.chan_id = 1;
	hunting.result_cb = NULL;

	/* set up caller data */

	if (strlen(called) > 0) {
		ftdm_set_string(call_data.dnis.digits, called);
	}
	call_data.call_id = 0;

	ftdm_call_place(&call_data, &hunting);

	return 0;
}

int sangoma_freetdm_channel_list(ftdm_span_t * ftdm_span, int channel)
{
	int idx = 0;
	int ret = 1;
	ftdm_channel_t *fchan;
	ftdm_alarm_flag_t alarmbits = FTDM_ALARM_NONE;
	ftdm_signaling_status_t sigstatus;

	fchan = ftdm_span_get_channel(ftdm_span, channel);
	ftdm_channel_get_alarms(fchan, &alarmbits);

	if ((FTDM_SUCCESS == ftdm_span_get_sig_status(ftdm_span, &sigstatus))) {
		ret = 0;
		ftdm_log(FTDM_LOG_DEBUG,
				 "\n+OK\n\n"
				 "\n***************************************************\n"
				 "span: %u (%s)\n"
				 "type: %s\n"
				 "physical_status: %s\n"
				 "signaling_status: %s\n"
				 "chan_count: %u\n"
				 "***************************************************\n",
				 idx,
				 ftdm_span_get_name(ftdm_span),
				 ftdm_span_get_trunk_type_str(ftdm_span),
				 alarmbits ? "alarmed" : "ok",
				 ftdm_signaling_status2str(sigstatus), ftdm_span_get_chan_count(ftdm_span));
	} else {
		ret = 0;
		ftdm_log(FTDM_LOG_DEBUG,
				 "\n+OK\n\n"
				 "\n***************************************************\n"
				 "span: %u (%s)\n"
				 "type: %s\n"
				 "physical_status: %s\n"
				 "chan_count: %u\n"
				 "***************************************************\n",
				 idx,
				 ftdm_span_get_name(ftdm_span),
				 ftdm_span_get_trunk_type_str(ftdm_span),
				 alarmbits ? "alarmed" : "ok", ftdm_span_get_chan_count(ftdm_span));
	}
	return ret;
}

int sangoma_freetdm_gsm_span_status(ftdm_span_t * ftdm_span)
{
	char cmd[512];
	char *rply;

	sprintf(cmd, "gsm status %s", ftdm_span_get_name(ftdm_span));

	rply = ftdm_api_execute(cmd);
	if (rply) {
		ftdm_log(FTDM_LOG_DEBUG, "%s", rply);
		ftdm_free(rply);
	}

	return 0;
}

int sangoma_freetdm_channel_dump(ftdm_span_t * ftdm_span, int channel)
{
	int idx = 0;
	ftdm_alarm_flag_t alarmbits = FTDM_ALARM_NONE;
	ftdm_signaling_status_t sigstatus;
	ftdm_channel_t *fchan;

	fchan = ftdm_span_get_channel(ftdm_span, channel);
	ftdm_channel_get_alarms(fchan, &alarmbits);

	if ((FTDM_SUCCESS == ftdm_span_get_sig_status(ftdm_span, &sigstatus))) {
		ftdm_log(FTDM_LOG_DEBUG,
				 "\n+OK\n\n"
				 "\n***************************************************\n"
				 "span: %u (%s)\n"
				 "type: %s\n"
				 "physical_status: %s\n"
				 "signaling_status: %s\n"
				 "chan_count: %u\n"
				 "***************************************************\n",
				 idx,
				 ftdm_span_get_name(ftdm_span),
				 ftdm_span_get_trunk_type_str(ftdm_span),
				 alarmbits ? "alarmed" : "ok",
				 ftdm_signaling_status2str(sigstatus), ftdm_span_get_chan_count(ftdm_span));

	} else {
		ftdm_log(FTDM_LOG_DEBUG,
				 "\n+OK\n"
				 "\n***************************************************\n"
				 "span: %u (%s)\n"
				 "type: %s\n"
				 "physical_status: %s\n"
				 "chan_count: %u\n"
				 "***************************************************\n",
				 idx,
				 ftdm_span_get_name(ftdm_span),
				 ftdm_span_get_trunk_type_str(ftdm_span),
				 alarmbits ? "alarmed" : "ok", ftdm_span_get_chan_count(ftdm_span));
	}

	return 0;
}

/*===================================================
 * ISDN TAP CLI COMMAND
 *==================================================*/
int sangoma_freetdm_isdn_tap_statistics(void)
{
	char cmd[512];
	char *rply;

	sprintf(cmd, "sangoma_isdn_tap statistics");

	rply = ftdm_api_execute(cmd);
	if (rply) {
		ftdm_log(FTDM_LOG_DEBUG, "\n%s", rply);
		ftdm_free(rply);
	}

	return 0;
}

/*===================================================
 * Private Configuration Functions
 *==================================================*/

/* FreeTDM gsm specific span create */
static int sangoma_freetdm_gsm_create_span(sangoma_port_obj_t port, ftdm_span_t * span)
{
	uint32_t chan_configured = 0;
	int configured = 0;
	char wp_bchans[255];
	char wp_dchans[255];
	ftdm_channel_config_t chan_config;
	int port_id = sangoma_config_port_get_port_id(port);

	memset(wp_bchans, 0, sizeof(wp_bchans));
	memset(wp_dchans, 0, sizeof(wp_dchans));
	memset(&chan_config, 0, sizeof(chan_config));

	/* Setting span trunk type */
	ftdm_span_set_trunk_type(span, FTDM_TRUNK_GSM);

	snprintf(wp_bchans, sizeof(wp_dchans), "%d:1", port_id);
	ftdm_configure_span_channels(span, wp_bchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u B channels on port %i\n",
			 ftdm_span_get_name(span), chan_configured, port_id);
	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure B channels\n");
		return configured;
	}

	configured += chan_configured;

	/* Configuring span d-channel */
	chan_configured = 0;

	chan_config.type = FTDM_CHAN_TYPE_DQ921;
	snprintf(wp_dchans, sizeof(wp_dchans), "%d:2", port_id);

	ftdm_configure_span_channels(span, wp_dchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u D channels\n", ftdm_span_get_name(span), chan_configured);

	configured += chan_configured;

	return configured;

}

/* FreeTDM tapping specific span create */
static int sangoma_freetdm_tapping_create_pri_span(sangoma_port_obj_t port, ftdm_span_t * span)
{
	uint32_t chan_configured = 0;
	int configured = 0;
	char wp_bchans[255];
	char wp_dchans[255];
	ftdm_channel_config_t chan_config;
	int port_id = sangoma_config_port_get_port_id(port);

	memset(wp_bchans, 0, sizeof(wp_bchans));
	memset(wp_dchans, 0, sizeof(wp_dchans));
	memset(&chan_config, 0, sizeof(chan_config));

	/* Setting span trunk type */
	ftdm_span_set_trunk_type(span, FTDM_TRUNK_E1);

	snprintf(wp_bchans, sizeof(wp_dchans), "%d:1-15", port_id);
	ftdm_configure_span_channels(span, wp_bchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u B channels on port %i\n",
			 ftdm_span_get_name(span), chan_configured, port_id);
	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure B channels [1-15] on %s\n", ftdm_span_get_name(span));
		return configured;
	}

	configured += chan_configured;
	chan_configured = 0;

	snprintf(wp_bchans, sizeof(wp_dchans), "%d:17-31", port_id);
	ftdm_configure_span_channels(span, wp_bchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u B channels on port %i\n",
			 ftdm_span_get_name(span), chan_configured, port_id);
	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure B channels [15-31] on %s\n", ftdm_span_get_name(span));
		return configured;
	}

	configured += chan_configured;
	chan_configured = 0;

	/* Configuring span d-channel */
	chan_config.type = FTDM_CHAN_TYPE_DQ921;
	snprintf(wp_dchans, sizeof(wp_dchans), "%d:16", port_id);

	ftdm_configure_span_channels(span, wp_dchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u D channels\n", ftdm_span_get_name(span), chan_configured);
	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure D channel 16 on %s\n", ftdm_span_get_name(span));
		return configured;
	}

	configured += chan_configured;

	return configured;
}

static int sangoma_freetdm_tapping_create_bri_span(sangoma_port_obj_t port, ftdm_span_t * span)
{
	uint32_t chan_configured = 0;
	int configured = 0;
	char wp_bchans[255];
	char wp_dchans[255];
	ftdm_channel_config_t chan_config;
	int port_id = sangoma_config_port_get_port_id(port);

	memset(wp_bchans, 0, sizeof(wp_bchans));
	memset(wp_dchans, 0, sizeof(wp_dchans));
	memset(&chan_config, 0, sizeof(chan_config));

	/* Setting span trunk type */
	ftdm_span_set_trunk_type(span, FTDM_TRUNK_BRI);

	snprintf(wp_bchans, sizeof(wp_dchans), "%d:1-2", port_id);
	ftdm_configure_span_channels(span, wp_bchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u B channels on port %i\n",
			 ftdm_span_get_name(span), chan_configured, port_id);
	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure B channels [1-2] on %s\n", ftdm_span_get_name(span));
		return -1;
	}

	configured += chan_configured;

	/* Configuring span d-channel */
	chan_configured = 0;

	chan_config.type = FTDM_CHAN_TYPE_DQ921;
	snprintf(wp_dchans, sizeof(wp_dchans), "%d:3", port_id);

	ftdm_configure_span_channels(span, wp_dchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u D channels\n", ftdm_span_get_name(span), chan_configured);
	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure D channel 3 on %s\n", ftdm_span_get_name(span));
		return -1;
	}

	configured += chan_configured;

	return configured;
}

/* FreeTDM isup signalling specific span create */
static int sangoma_freetdm_tapping_create_isup_sig_span(sangoma_port_obj_t port, ftdm_span_t * span)
{
	uint32_t chan_configured = 0;
	int configured = 0;
	char wp_bchans[255];
	char wp_dchans[255];
	ftdm_channel_config_t chan_config;
	int port_id = sangoma_config_port_get_port_id(port);

	memset(wp_bchans, 0, sizeof(wp_bchans));
	memset(wp_dchans, 0, sizeof(wp_dchans));
	memset(&chan_config, 0, sizeof(chan_config));

	/* Setting span trunk type */
	ftdm_span_set_trunk_type(span, FTDM_TRUNK_E1);

	snprintf(wp_bchans, sizeof(wp_dchans), "%d:1-15", port_id);
	ftdm_configure_span_channels(span, wp_bchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u B channels on port %i\n",
			 ftdm_span_get_name(span), chan_configured, port_id);
	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure B channels [1-15] on %s\n", ftdm_span_get_name(span));
		return configured;
	}

	configured += chan_configured;
	chan_configured = 0;

	snprintf(wp_bchans, sizeof(wp_dchans), "%d:17-31", port_id);
	ftdm_configure_span_channels(span, wp_bchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u B channels on port %i\n",
			 ftdm_span_get_name(span), chan_configured, port_id);
	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure B channels [17-31] on %s\n", ftdm_span_get_name(span));
		return configured;
	}

	configured += chan_configured;

	/* Configuring span d-channel i.e. signalling in terms of SS7 */
	chan_configured = 0;

	chan_config.type = FTDM_CHAN_TYPE_DQ921;
	snprintf(wp_dchans, sizeof(wp_dchans), "%d:16", port_id);

	ftdm_configure_span_channels(span, wp_dchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u D channels\n", ftdm_span_get_name(span), chan_configured);
	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure D channel 16 on %s\n", ftdm_span_get_name(span));
		return -1;
	}

	configured += chan_configured;
	return configured;
}

/* FreeTDM isup pure voice specific span create */
static int sangoma_freetdm_tapping_create_isup_pure_voice_span(sangoma_port_obj_t port, ftdm_span_t * span)
{
	uint32_t chan_configured = 0;
	int configured = 0;
	char wp_bchans[255];
	char wp_dchans[255];
	ftdm_channel_config_t chan_config;
	int port_id = sangoma_config_port_get_port_id(port);

	memset(wp_bchans, 0, sizeof(wp_bchans));
	memset(wp_dchans, 0, sizeof(wp_dchans));
	memset(&chan_config, 0, sizeof(chan_config));

	/* Setting span trunk type */
	ftdm_span_set_trunk_type(span, FTDM_TRUNK_E1);

	snprintf(wp_bchans, sizeof(wp_dchans), "%d:1-31", port_id);
	ftdm_configure_span_channels(span, wp_bchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u B channels on port %i\n",
			 ftdm_span_get_name(span), chan_configured, port_id);
	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure B channels [1-15] on %s\n", ftdm_span_get_name(span));
		return configured;
	}

	configured += chan_configured;
	return configured;
}

/* FreeTDM mtp2 api specific span create */
static int sangoma_freetdm_mtp2_api_create_span(sangoma_port_obj_t port, ftdm_span_t * span)
{
	uint32_t chan_configured = 0;
	int configured = 0;
	char wp_bchans[255];
	char wp_dchans[255];
	ftdm_channel_config_t chan_config;
	int port_id = sangoma_config_port_get_port_id(port);

	memset(wp_bchans, 0, sizeof(wp_bchans));
	memset(wp_dchans, 0, sizeof(wp_dchans));
	memset(&chan_config, 0, sizeof(chan_config));

	/* Setting span trunk type */
	ftdm_span_set_trunk_type(span, FTDM_TRUNK_E1);

	chan_config.type = FTDM_CHAN_TYPE_DQ921;
	snprintf(wp_dchans, sizeof(wp_dchans), "%d:1", port_id);

	ftdm_configure_span_channels(span, wp_dchans, &chan_config, &chan_configured);
	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u D channels\n", ftdm_span_get_name(span), chan_configured);
	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure D channel 1 on %s\n", ftdm_span_get_name(span));
		return -1;
	}

	configured += chan_configured;
	return configured;
}

/* FreeTDM gsm specific configuration */
static int sangoma_freetdm_gsm_config(sangoma_port_obj_t port, ftdm_span_t * span, span_config_t * span_conf)
{
	unsigned paramindex = 0;
	ftdm_conf_parameter_t spanparameters[30];

	/* Set moduleType type */
	spanparameters[paramindex].var = "moduleType";
	spanparameters[paramindex].val = "telit";
	paramindex++;

	/* The last parameter .var member must be NULL! */
	spanparameters[paramindex].var = NULL;
	spanparameters[paramindex].val = NULL;

	/* Configure span */
	if (ftdm_configure_span_signaling(span, "gsm", on_freetdm_event_signal, spanparameters) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Error configuring Sangoma GSM FreeTDM span %d\n", ftdm_span_get_id(span));
		return -1;
	}

	return 0;
}

/* FreeTDM tapping specific configuration */
static int sangoma_freetdm_tapping_config(sangoma_port_obj_t port, ftdm_span_t * span, span_config_t * span_conf)
{
	unsigned paramindex = 0;

	ftdm_conf_parameter_t spanparameters[30];

	/* Set Mixedaudio Type */
	if (strlen(span_conf->config.isdn.mixaudio_mode) != 0) {
		spanparameters[paramindex].var = "mixaudio";
		spanparameters[paramindex].val = span_conf->config.isdn.mixaudio_mode;
		paramindex++;
	}

	/* Set span interface */
	if (strlen(span_conf->config.isdn.interface_name) != 0) {
		spanparameters[paramindex].var = "interface";
		spanparameters[paramindex].val = span_conf->config.isdn.interface_name;
		paramindex++;
	}

	/* Set peer span */
	if (strlen(span_conf->config.isdn.peer_span) != 0) {
		spanparameters[paramindex].var = "peerspan";
		spanparameters[paramindex].val = span_conf->config.isdn.peer_span;
		paramindex++;
	}

	/* Set Span interface type */
	if (strlen(span_conf->config.isdn.interface_type) != 0) {
		spanparameters[paramindex].var = "interface_type";
		spanparameters[paramindex].val = span_conf->config.isdn.interface_type;
		paramindex++;
	}

	/* Set Span switch type */
	if (strlen(span_conf->config.isdn.switch_type) != 0) {
		spanparameters[paramindex].var = "switch_type";
		spanparameters[paramindex].val = span_conf->config.isdn.switch_type;
		paramindex++;
	}

	/* Set DEBUG type */
	if (strlen(span_conf->config.isdn.debug) != 0) {
		spanparameters[paramindex].var = "debug";
		spanparameters[paramindex].val = span_conf->config.isdn.debug;
		paramindex++;
	}

	/* The last parameter .var member must be NULL! */
	spanparameters[paramindex].var = NULL;
	spanparameters[paramindex].val = NULL;

	/* Configure span */
	if (ftdm_configure_span_signaling(span, "sangoma_isdn_tap", on_freetdm_event_signal, spanparameters) !=
		FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Error configuring Sangoma ISDN Tapping span %d\n", ftdm_span_get_id(span));
		return -1;
	}

	return 0;
}

/* FreeTDM isup tapping specific configuration */
static int sangoma_freetdm_isup_tapping_config(sangoma_port_obj_t port, ftdm_span_t * span, span_config_t * span_conf)
{
	unsigned paramindex = 0;

	ftdm_conf_parameter_t spanparameters[30];

	/* Set Mixedaudio Type */
	if (strlen(span_conf->config.isup.mixaudio_mode) != 0) {
		spanparameters[paramindex].var = "mixaudio";
		spanparameters[paramindex].val = span_conf->config.isup.mixaudio_mode;
		paramindex++;
	}


	/* Set peer span */
	if (strlen(span_conf->config.isup.peer_span) != 0) {
		spanparameters[paramindex].var = "peerspan";
		spanparameters[paramindex].val = span_conf->config.isup.peer_span;
		paramindex++;
	}

	/* Set Span switch type */
	if (strlen(span_conf->config.isup.switch_type) != 0) {
		spanparameters[paramindex].var = "switch_type";
		spanparameters[paramindex].val = span_conf->config.isup.switch_type;
		paramindex++;
	}

	/* Set Span Cicbase type */
	if (strlen(span_conf->config.isup.cicbase) != 0) {
		spanparameters[paramindex].var = "cicbase";
		spanparameters[paramindex].val = span_conf->config.isup.cicbase;
		paramindex++;
	}

	/* Set whether span is pure voice or signalling type */
	if (strlen(span_conf->config.isup.is_voice) != 0) {
		spanparameters[paramindex].var = "is_voice";
		spanparameters[paramindex].val = span_conf->config.isup.is_voice;
		paramindex++;
	}

	/* Set is there is any voice span w.r.t this signalling span */
	if (strlen(span_conf->config.isup.voice_span) != 0) {
		spanparameters[paramindex].var = "voice_span";
		spanparameters[paramindex].val = span_conf->config.isup.voice_span;
		paramindex++;
	}

	/* The last parameter .var member must be NULL! */
	spanparameters[paramindex].var = NULL;
	spanparameters[paramindex].val = NULL;

	/* Configure span */
	if (ftdm_configure_span_signaling(span, "sangoma_isup_tap", on_freetdm_event_signal, spanparameters) !=
		FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Error configuring Sangoma ISUP Tapping span %d\n", ftdm_span_get_id(span));
		return -1;
	}

	return 0;
}

/* FreeTDM mtp2 api specific configuration */
static int sangoma_freetdm_mtp2_api_config(sangoma_port_obj_t port, ftdm_span_t * span, span_config_t *span_conf)
{
	unsigned paramindex = 0;
	int idx = 0;
	char *name = NULL;
	char *var = NULL;
	char *val = NULL;

	ftdm_conf_parameter_t spanparameters[30];

	if (strlen(span_conf->config.mtp2_api.span_name) == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Error getting span name\n");
		return -1;
	}
	name = span_conf->config.mtp2_api.span_name;
	
	if (!(span_conf->config.mtp2_api.ss7confnode)) {
		ftdm_log(FTDM_LOG_ERROR, "Error finding ss7config configured node for FreeTDM span: %s\n", name);
		return -1;
	}

	memset(spanparameters, 0, sizeof(spanparameters));
	paramindex = 0;
	spanparameters[paramindex].var = "confnode";
	spanparameters[paramindex].ptr = span_conf->config.mtp2_api.ss7confnode;
	paramindex++;

	for (idx = 0; idx < 10; idx++) {
		if (strlen(span_conf->config.mtp2_api.p_name[idx]) == 0) {
			break;
		}

		var = span_conf->config.mtp2_api.p_name[idx];
		val = span_conf->config.mtp2_api.p_value[idx];

		spanparameters[paramindex].var = var;
		spanparameters[paramindex].val = val;
		paramindex++;
	}
	/* The last parameter .var member must be NULL! */
	spanparameters[paramindex].var = NULL;
	spanparameters[paramindex].val = NULL;

	/* Configure span */
	if (ftdm_configure_span_signaling(span, "sangoma_ss7", on_freetdm_event_signal, spanparameters) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Error configuring ss7 FreeTDM span %d\n", ftdm_span_get_id(span));
		return -1;
	}

	return 0;
}

/*===================================================
 * Private Configuration Functions
 * Analog cards
 *==================================================*/

static int sangoma_freetdm_analog_create_span(sangoma_port_obj_t port, ftdm_span_t * span)
{
	uint32_t chan_configured = 0;
	int configured = 0;
	char wp_bchans[255];
	ftdm_channel_config_t chan_config;
	int port_id = sangoma_config_port_get_port_id(port);

	memset(wp_bchans, 0, sizeof(wp_bchans));
	memset(&chan_config, 0, sizeof(chan_config));
	chan_config.type = FTDM_CHAN_TYPE_FXS;

	/* Setting span trunk type */
	ftdm_span_set_trunk_type(span, FTDM_TRUNK_FXS);

	snprintf(wp_bchans, sizeof(wp_bchans), "%d:1", port_id);
	if (ftdm_configure_span_channels(span, wp_bchans, &chan_config, &chan_configured) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Cannot create FXO channel %s\n", wp_bchans);
		return configured;
	}

	ftdm_log(FTDM_LOG_DEBUG, "%s: ftdm configure %u FXO channels on port %i\n",
			 ftdm_span_get_name(span), chan_configured, port_id);

	if (chan_configured == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to configure FXO channel %s.\n", wp_bchans);
		return configured;
	}

	configured += chan_configured;
	return configured;

}

static int sangoma_freetdm_analog_config(sangoma_port_obj_t port, ftdm_span_t * span, span_config_t * span_conf)
{
	ftdm_conf_parameter_t spanparameters[10];
	unsigned paramindex = 0;

	spanparameters[paramindex].var = "tonemap";
	spanparameters[paramindex].val = "us";
	paramindex++;

	spanparameters[paramindex].var = "digit_timeout";
	spanparameters[paramindex].val = "0";
	paramindex++;

	spanparameters[paramindex].var = "enable_callerid";
	spanparameters[paramindex].val = "true";
	paramindex++;

	spanparameters[paramindex].var = NULL;
	spanparameters[paramindex].val = NULL;

	if (ftdm_configure_span_signaling(span, "analog_tap", on_freetdm_event_signal, spanparameters) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Error configuring Sangoma Analog FreeTDM span %d\n", ftdm_span_get_id(span));
	}

	return 0;
}

/******************* Logging module *******************************************/

const char* sangoma_freetdm_cut_path(const char *file_path)
{
        const char *p, *ret = file_path;
        const char delims[] = "/\\";
        const char *i;

        if (file_path) {
                for (i = delims; *i; i++) {
                        p = file_path;
                        while ((p = strchr(p, *i)) != 0) {
                                ret = ++p;
                        }
                }
                return ret;
        } else {
                return NULL;
        }
}

void sangoma_freetdm_logger(const char *file, const char *func, int line, int level, const char *fmt, ...)
{
	char data[1024] = { 0 };
	char curr_time[80] = { 0 };
	va_list ap;
	const char *filep = (file ? sangoma_freetdm_cut_path(file) : "");

	if (level < 0 || level > 7) {
		level = 7;
	}

	va_start(ap, fmt);

	vsnprintf(data, sizeof(data), fmt, ap);

	/* get current date and time */
	sangoma_get_timestamp_str(curr_time);

	fprintf(stderr, "%s [%s] %s:%d %s() %s", curr_time, SANGOMA_FTDM_LEVEL_NAMES[level], filep, line, func, data);

	va_end(ap);
}

void sangoma_freetdm_exec_io_stats(const char* cmd, ftdm_span_t * span)
{
	ftdm_iterator_t *chaniter = NULL;
	ftdm_iterator_t *curr = NULL;
	ftdm_channel_t *ftdmchan = NULL;

	if ((NULL == cmd) || (NULL == span)) {
		ftdm_log(FTDM_LOG_ERROR, "Error Enabling IO Stats..invalid span \n");
		return;
	}

	chaniter = ftdm_span_get_chan_iterator(span, NULL);
	for (curr = chaniter; curr; curr = ftdm_iterator_next(curr)) {
		ftdmchan = (ftdm_channel_t*)ftdm_iterator_current(curr);
		sangoma_freetdm_exec_io_command(cmd, ftdmchan);
	}
	ftdm_iterator_free(chaniter);
}

/* This API will be used to enable/disable IOSTATS on ftdm channels */
void sangoma_freetdm_exec_io_command(const char *cmd, ftdm_channel_t *fchan)
{
	int enable = 0;
	ftdm_channel_iostats_t stats;
	char buffer[1024] = {0};
	int len = 0;

	if (!fchan || !cmd)
		return;

	if (!strcasecmp("enable", cmd)) {
		enable = 1; 
		ftdm_channel_command(fchan, FTDM_COMMAND_SWITCH_IOSTATS, &enable);
	} else if (!strcasecmp("disable", cmd)) {
		enable = 0; 
		ftdm_channel_command(fchan, FTDM_COMMAND_SWITCH_IOSTATS, &enable);
	} else if (!strcasecmp("flush", cmd)) {
		ftdm_channel_command(fchan, FTDM_COMMAND_FLUSH_IOSTATS, NULL);
	} else {
		ftdm_channel_command(fchan, FTDM_COMMAND_GET_IOSTATS, &stats);
		len = len + sprintf(buffer, "-- IO statistics for channel %d:%d --\n",
				ftdm_channel_get_span_id(fchan), ftdm_channel_get_id(fchan));
		len = len + sprintf(buffer+len, "Rx errors: %u\n", stats.rx.errors);
		len = len + sprintf(buffer+len, "Rx queue size: %u\n", stats.rx.queue_size);
		len = len + sprintf(buffer+len, "Rx queue len: %u\n", stats.rx.queue_len);
		len = len + sprintf(buffer+len, "Rx count: %lu\n", stats.rx.packets);
		len = len + sprintf(buffer+len, "Tx errors: %u\n", stats.tx.errors);
		len = len + sprintf(buffer+len, "Tx queue size: %u\n", stats.tx.queue_size);
		len = len + sprintf(buffer+len, "Tx queue len: %u\n", stats.tx.queue_len);
		len = len + sprintf(buffer+len, "Tx count: %lu\n", stats.tx.packets);
		len = len + sprintf(buffer+len, "Tx idle: %lu\n", stats.tx.idle_packets);

		ftdm_log(FTDM_LOG_DEBUG, "%s\n", buffer);

		//fprintf(stderr,"%s\n", buffer);
		ftdm_log(FTDM_LOG_DEBUG, "%s\n",buffer);
	}
}

void sangoma_freetdm_set_span_sig_queuesize(ftdm_span_t * span, int rxsize)
{
	ftdm_iterator_t *chaniter = NULL;
	ftdm_iterator_t *curr = NULL;
	ftdm_channel_t *ftdmchan = NULL;

	if (!rxsize || (NULL == span)) {
		ftdm_log(FTDM_LOG_ERROR, "Error Setting RxQueue size ..invalid span/rxsize \n");
		return;
	}

	chaniter = ftdm_span_get_chan_iterator(span, NULL);
	for (curr = chaniter; curr; curr = ftdm_iterator_next(curr)) {
		ftdmchan = (ftdm_channel_t*)ftdm_iterator_current(curr);
		if (SNG_FTDM_IS_DCHAN(ftdmchan))
			sangoma_freetdm_set_chan_sig_rx_queuesize(ftdmchan, rxsize);
	}
	ftdm_iterator_free(chaniter);
}

void sangoma_freetdm_set_chan_sig_rx_queuesize(ftdm_channel_t *fchan, int rxsize)
{
	ftdm_status_t   ret_status;

	if (!fchan) return;

	if (!rxsize) return;

	if (SNG_FTDM_IS_DCHAN(fchan)) {

		fprintf(stderr,"Setting Rx Queue size for channel %d:%d --\n", ftdm_channel_get_span_id(fchan), ftdm_channel_get_id(fchan));

		ret_status = ftdm_channel_command(fchan, FTDM_COMMAND_SET_RX_QUEUE_SIZE, &rxsize);

		if (ret_status != FTDM_SUCCESS) {
			ftdm_log(FTDM_LOG_ERROR, "Failed to set Rx Queue size[%d]", rxsize);
		}
	}
}

/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4:
 */
