#ifndef __LIBSANGOMA_HW_CONFIG_H__
#define __LIBSANGOMA_HW_CONFIG_H__

#include "libsangoma_tdm_sdk_version.h"
#include "libsangoma.h"

/*!
  \typedef sangoma_port_obj_t
  \brief Sangoma port object structure

  Sangoma port object structure contains all hw port information
  It also contains all hw configuration structures needed to
  operate the port. Start,Stop,Save Config etc..

  The object is provided to user opaque in order to simplify development.
  Methods are provided to access and manipulate the port object.
*/
typedef void *sangoma_port_obj_t;

/*!
  \ Media type enums 
*/
typedef enum {
	SNG_WAN_MEDIA_NONE,   
	SNG_WAN_MEDIA_T1,    
	SNG_WAN_MEDIA_E1,   
	SNG_WAN_MEDIA_56K,
	SNG_WAN_MEDIA_DS3, 
	SNG_WAN_MEDIA_E3,  
	SNG_WAN_MEDIA_STS1, 
	SNG_WAN_MEDIA_J1,  
	SNG_WAN_MEDIA_FXOFXS, 
	SNG_WAN_MEDIA_BRI,   
	SNG_WAN_MEDIA_SERIAL,
	SNG_WAN_MEDIA_GSM,  
}sng_media_type_e;

/*!
  \ Line code type enum
*/
typedef enum {
	SNG_WAN_LCODE_NONE,
	SNG_WAN_LCODE_AMI,
	SNG_WAN_LCODE_B8ZS,
	SNG_WAN_LCODE_HDB3,
	SNG_WAN_LCODE_B3ZS,
}sng_line_code_types_e;

/*!
  \ Framing code type enum
*/
typedef enum {
	SNG_WAN_FR_NONE,           /* 0 */ 
	SNG_WAN_FR_ESF,            /* 1 */ 	
	SNG_WAN_FR_D4,             /* 2 */ 
	SNG_WAN_FR_ESF_JAPAN,      /* 3 */ 
	SNG_WAN_FR_CRC4,           /* 4 */ 
	SNG_WAN_FR_NCRC4,          /* 5 */ 
	SNG_WAN_FR_UNFRAMED,        
	SNG_WAN_FR_E3_G751,         
	SNG_WAN_FR_E3_G832,         
	SNG_WAN_FR_DS3_Cbit,        
	SNG_WAN_FR_DS3_M13,         
	SNG_WAN_FR_SLC96,
}sng_frame_modes_e;           

/*!
  \typedef sng_wp_config_t
  \brief Sangoma SDK Application structure to fill wanpipe/port low level configuration parameters
*/
typedef struct sng_wp_config {
	int wp_config_pres;
	sng_media_type_e media;
	sng_frame_modes_e frame_mode; /* CRC4/NCRC4 etc */
	sng_line_code_types_e line_code;  /* HDB3  etc */
	int high_impedance_mode; /* High Impendence on/off */
}sng_wp_config_t;


/*!
  \typedef sangoma_hw_config_reg_t
  \brief Sangoma operations callback structure

  The callback structure allows user application to pass on
  use specific information to the sangoma library.
  
  user_obj - is a user pointer that will be passed back to the user application
             on each callback.
  
  hw_probe_callback() - is functin that will be called for each port detected
                        during hwprobe. 
  hw_probe_callback_post() - is functin that will be called for each port detected
                        after the initial hw_probe_callback.  
                        
  Depending on use case scenario, user can configure all ports first in hw_probe_callback()
  and then start all ports in hw_probe_callback_post() 
 
*/
typedef struct sangoma_hw_config_reg {
	sangoma_port_obj_t user_obj;
	int (*hwprobe_callback) (void *sangoma_port_obj, int port_id, void *user_obj);
	int (*hwprobe_callback_post) (void *sangoma_port_obj, int port_id, void *user_obj);
} sangoma_hw_config_reg_t;

/* Exported Functions */

/*!
  \fn int sangoma_config_reg (sangoma_hw_config_reg_t *reg)
  \brief Register hw callback functions into libsangoma
  \param reg callback registration structure 
  \return File Descriptor: -1 error, 0 success

  A very first access to the sangoma operations library.
  User application registers hw callback functions.
  These functions will be called on every hw port detect. 
*/
int _LIBSNG_CALL sangoma_config_reg(sangoma_hw_config_reg_t * reg);

/*!
  \fn int sangoma_set_wp_config(sangoma_port_obj_t sangoma_port_obj,sng_wp_config_t* app_cfg); 
  \brief Set hw port configuration option via application 
  \param sangoma_port_obj sangoma hw port object, contains all hw and cfg information 
  \param sng_wp_config_t sangoma application hw cfg interface structure

  
  This function will save the user  defined HW configuration to low level port 
*/
int _LIBSNG_CALL sangoma_set_wp_config(sangoma_port_obj_t sangoma_port_obj, sng_wp_config_t* app_cfg);


/*!
  \fn int sangoma_config_port_cfg_set_option(sangoma_port_obj_t sangoma_port_obj, char *key, char *value)
  \brief Set hw port configuration option via key value pair
  \param sangoma_port_obj sangoma hw port object, contains all hw and cfg information 
  \param key configuration option
  \param value configuration option value
  \return File Descriptor: -1 error, 0 success

  This is a method that operates on sangoma_port_obj.
  Set a configuration option via key value ascii pair.
  Can be run multiple time on same object to configure each configuration option.

*/
int _LIBSNG_CALL sangoma_config_port_cfg_set_option(sangoma_port_obj_t sangoma_port_obj, char *key, char *value);

/*!
  \fn int sangoma_config_port_cfg_get_option(sangoma_port_obj_t sangoma_port_obj, char *key, char *value)
  \brief Get hw port configuration option via key value pair
  \param sangoma_port_obj sangoma hw port object, contains all hw and cfg information 
  \param key configuration option
  \param value port object current value of for the key
  \return File Descriptor: -1 error, 0 success

  This is a method that operates on sangoma_port_obj.
  Get a configuration option value that is currently configured from port obj.
  Can be run multiple time on same object to retrieve configuratin options.
*/
int _LIBSNG_CALL sangoma_config_port_cfg_get_option(sangoma_port_obj_t sangoma_port_obj, char *key, char *value);

int _LIBSNG_CALL sangoma_config_port_cfg_dump(sangoma_port_obj_t sangoma_port_obj);
int _LIBSNG_CALL sangoma_config_port_hwinfo_dump(sangoma_port_obj_t sangoma_port_obj);

/*!
  \fn char* sangoma_config_port_hwinfo_get_model_str(sangoma_port_obj_t sangoma_port_obj)
  \brief Get hw type name from the port object
  \param sangoma_port_obj sangoma hw port object, contains all hw and cfg information 
  \return File Descriptor: NULL is error, Valid String is success

  This is a method that operates on sangoma_port_obj.
  Get hw type name from the port object
*/
char *sangoma_config_port_hwinfo_get_model_str(sangoma_port_obj_t sangoma_port_obj);

/*!
  \fn unsigned int sangoma_config_port_hwinfo_get_model_id(sangoma_port_obj_t sangoma_port_obj)
  \brief Get hw type id from the port object
  \param sangoma_port_obj sangoma hw port object, contains all hw and cfg information
  \return File Descriptor: -1 is error, 0 or greater is success

  This is a method that operates on sangoma_port_obj.
  Get hw type id from the port object
*/
unsigned int _LIBSNG_CALL sangoma_config_port_hwinfo_get_model_id(sangoma_port_obj_t sangoma_port_obj);
unsigned int _LIBSNG_CALL sangoma_config_port_hwinfo_get_port_number(sangoma_port_obj_t sangoma_port_obj);
unsigned int _LIBSNG_CALL sangoma_config_port_hwinfo_get_pcislot(sangoma_port_obj_t sangoma_port_obj);
unsigned int _LIBSNG_CALL sangoma_config_port_hwinfo_get_pcibus(sangoma_port_obj_t sangoma_port_obj);
unsigned int _LIBSNG_CALL sangoma_config_port_hwinfo_get_fimrware_version(sangoma_port_obj_t sangoma_port_obj);
unsigned int _LIBSNG_CALL sangoma_config_port_get_port_id(sangoma_port_obj_t sangoma_port_obj);

/*!
  \fn int sangoma_config_port_save(sangoma_port_obj_t sangoma_port_obj)
  \brief Save current configuration in port obj to driver. Driver restart is needed to apply
  \param sangoma_port_obj sangoma hw port object, contains all hw and cfg information
  \return File Descriptor: Non Zero is error, 0  is success

  This is a method that operates on sangoma_port_obj.
  Save current configuration in port obj to driver. Driver restart is needed to apply.
*/
int _LIBSNG_CALL sangoma_config_port_save(sangoma_port_obj_t sangoma_port_obj);

/*!
  \fn int sangoma_config_port_apply(sangoma_port_obj_t sangoma_port_obj)
  \brief Reserved for future use
  \param sangoma_port_obj sangoma hw port object, contains all hw and cfg information
  \return File Descriptor: Non Zero is error, 0  is success

  Reserved for future use.
*/
int _LIBSNG_CALL sangoma_config_port_apply(sangoma_port_obj_t sangoma_port_obj);

/*!
  \fn int sangoma_config_port_start(sangoma_port_obj_t sangoma_port_obj)
  \brief Start the device driver/hw port
  \param sangoma_port_obj sangoma hw port object, contains all hw and cfg information
  \return File Descriptor: Non Zero is error, 0  is success

  This is a method that operates on sangoma_port_obj.
  Start the device driver/hw port. 
  The saved configuration will be applied to the port.
  At this point the port will try to connect to the line.
  LED on boards will become active.
  All interrupts are enabled. 
*/
int _LIBSNG_CALL sangoma_config_port_start(sangoma_port_obj_t sangoma_port_obj);

/*!
  \fn int sangoma_set_default_isup_pure_voice_span_config(sangoma_port_obj_t sangoma_port_obj)
  \brief Configure the device driver/hw port for pure voice channels.
  \param sangoma_port_obj sangoma hw port object, contains all hw and cfg information
  \return File Descriptor: Non Zero is error, 0  is success

  This is a method that operates on sangoma_port_obj.
  Configure the device driver/hw port for pure voice channels. 
  The saved configuration will be applied to the port.
  At this point the port will try to connect to the line.
*/
int _LIBSNG_CALL sangoma_set_default_isup_pure_voice_span_config(sangoma_port_obj_t sangoma_port_obj);

/*!
  \fn int sangoma_config_port_stop(sangoma_port_obj_t sangoma_port_obj)
  \brief Stop the device driver/hw port
  \param sangoma_port_obj sangoma hw port object, contains all hw and cfg information
  \return File Descriptor: Non Zero is error, 0  is success

  This is a method that operates on sangoma_port_obj.
  Stop the device driver/hw port.
  Driver will unconfigure the hwport. Interrutps are disabled.
*/
int _LIBSNG_CALL sangoma_config_port_stop(sangoma_port_obj_t sangoma_port_obj);

/*!
  \fn int sangoma_config_hwprobe(int *detected, int *configured)
  \brief Detect sangoma hardware on a machine
  \param detected detected will contain the number of ports detected on machine
  \param configured configured will contain the number of ports that were successfully configured via the callback function.
  \param operating_mode will contain the mode on which SDK application is operating i.e. ANALOG/ISUP_SIG/ISUP_VOICE/ISDN/MTP2_API_LSL/MTP2_API_HSL
  \return File Descriptor: Non Zero is error, 0  is success

  Detect sangoma hardware on a machine
  For each port a hw callback function will be called into the user application.
  The callback function will be give the sangoma _port_obj_t
     The port obj structure will contain all hw information about the port
     It will also contain fully initialized default configuration for the port based on port type.
     Meaning that user only needs to configure basic set of paramters before starting the port.
     Usually default configuration is enough to start the port successfully.

  In callback function user application should:
     Save the port object pointer in users data structure for later use.
     Configure the port obj via sangoma_config_port_cfg_set_option method for each cofiguration optin.
     Start sangoma hw port via sangoma_config_port_start

  In callback post function user applicaitn should:
     Create freetdm span
     Start freetdm span
*/
int _LIBSNG_CALL sangoma_config_hwprobe(int *detected, int *configured, int operating_mode);
int _LIBSNG_CALL sangoma_stop_all_wanpipe(void);

#ifdef __WINDOWS__
#include <ftdm_declare.h>
typedef struct ftdm_timezone {
	int tz_minuteswest;			/* minutes W of Greenwich */
	int tz_dsttime;				/* type of dst correction */
};
static inline int gettimeofday(struct timeval *tv, struct ftdm_timezone *tz)
{
	FILETIME ft;
	unsigned __int64 tmpres = 0;
	static int tzflag;
	if (NULL != tv) {
		GetSystemTimeAsFileTime(&ft);
		tmpres |= ft.dwHighDateTime;
		tmpres <<= 32;
		tmpres |= ft.dwLowDateTime;

		/*converting file time to unix epoch */
		tmpres /= 10;			/*convert into microseconds */
		tmpres -= DELTA_EPOCH_IN_MICROSECS;
		tv->tv_sec = (long)(tmpres / 1000000UL);
		tv->tv_usec = (long)(tmpres % 1000000UL);
	}
	if (NULL != tz) {
		if (!tzflag) {
			_tzset();
			tzflag++;
		}
		tz->tz_minuteswest = _timezone / 60;
		tz->tz_dsttime = _daylight;
	}
	return 0;
}
#endif

static inline void sangoma_get_timestamp_str(char *timestamp)
{
#ifdef __WINDOWS__
	__time64_t tv;
#else
	struct timeval tv;
#endif
	struct tm *ptm;
	char time_string[128];
	long milliseconds;

	if (!timestamp) {
		return;
	}

	/* Obtain the time of day, and convert it to a tm struct.  */
#ifdef __WINDOWS__
	_time64 (&tv);
	ptm = _localtime64(&tv);
#else
	gettimeofday(&tv, NULL);
	ptm = localtime(&tv.tv_sec);
#endif
	/* Format the date and time, down to a single second.  */
	strftime(time_string, sizeof(time_string), "%Y-%m-%d-%H-%M-%S", ptm);
	/* Compute milliseconds from microseconds.  */
#ifdef __WINDOWS__
	milliseconds = 0;
#else
	milliseconds = tv.tv_usec / 1000;
#endif
	/* Print the formatted time, in seconds, followed by a decimal point
	 * and the milliseconds.  */
	snprintf(timestamp, 80, "%s.%03ld", time_string, milliseconds);

	return;
}

/*!
  Defines all the operating mode in which SDK can operate
*/
typedef enum {
	SNG_ANALOG_TAPPING_MODE = 1,
	SNG_ISDN_PRI_TAPPING_MODE,
	SNG_ISDN_BRI_TAPPING_MODE,
	SNG_SS7_ISUP_SIG_TAPPING_MODE,
	SNG_SS7_ISUP_PURE_VOICE_TAPPING_MODE,
	SNG_SS7_MTP2_API_MODE_LSL,
	SNG_SS7_MTP2_API_MODE_HSL
} sng_tapping_mode_t;

#endif
